// $Id: DRAWxtl3.cxx 615 2007-01-12 17:36:59Z larry $
//
// module drawxtl3.cxx - part of DRAWxtl V5.3
// Coded using the FLTK 1.1.6 widget set
//
//     Larry W. Finger, Martin Kroeker and Brian Toby
//
//
// This module contains the following routines:
//
//  build_poly_list - generate the list of polygon vertices that describe the polyhedron
//  draw_cell - draw the unit-cell boundaries
//  draw_GL_triple - draw the axes for the vector triple in openGL
//  generate_spheres - output the list of spheres for the display lists 
//  generate_triple - generate the vertex locations for the triple vector
//  push_cylinder - routine to push an openGL cylinder that represents a unit-cell edge
//  generate_lsq_planes - generate best fitting planes 
//

#include "drawxtl.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <FL/glut.H>
#if defined(__APPLE__) 
#  include <openGL/glu.h>
#else
#  include <GL/glu.h>
#endif
#include "draw_ext.h"

#include "DRAWxtl_proto.h"

/* ************************************************************** */
/* ************************************************************** */

void build_poly_list (int numb_list, int polyno, int i0, int i1, float cx[3])

/* routine to take list of 'numb_list' indices in 's_vert'
   and find each plane on surface of convex polyhedron.  It then builds
   a list of pointers to these vertices for drawing program.  The algorithm
   is as follows:
   
   1.  If there are 4 vertices, the polyhedron is a tetrahedron and all
       combinations of 3 vertices constitute a face.

   2.  For more than 4 vertices, some planes will be interior.  These are
       found by choosing three points (indicated by i,j,k) and forming two
       vectors defined by V1 = P(j) - P(i) and V2 = P(k) - P(i).  For each
       additional point (l), a third vector V3 = P(l) - P(i) is formed and
       the volume enclosed by these three vectors V = V3 . V1 x V2 is
       calculated.  If this volume is zero, point l lies in the plane.  If
       the plane is on the surface of the polyhedron, all non-zero volumes
       must have the same sign, which may be positive or negative.
*/
/*  numb_list - number of candidate vertices */
/*  polyno - number of polygon - for polygon color */
/*  i0 - pointer to central atom */
/*  i1 - pointer to an atom in the polyhedron */
/*  cx - orthogonal coordinates of center of polyhedron */

{
  int i, j, k, l, m, n;                            /* local loop counters */
  float v0[3], v1[3], v2[3], v3[3], v4[3];      /* vectors used in calculation */
  float p[3];                     /* coordinates of point i */
  float vol, sign_z;
  int ind;                     /* indicator of surface plane */
  int *no;                     /* place to keep vertex id's */
  int imk, jmk, kmk;
  int ii, jj, ll;
  int jplane;
  double phi, cosphi;
  float xm, ym, zm, vx, vy, vz, bm, phi0, vxs, vys, vzs, bms;
  float d1, d2, d3, radius;
  int *ns;
  char edgecolor[40];
  static int warn = 0;
  float glr=0.0f, glg=0.0f, glb=0.0f;
  float Vol = 0.0f;             /* volume of polyhedron */
  float Area, Z0, Z;
  int ih[3];
  float sumth = 0.0f, sumth2 = 0.0f, sumd2 = 0.0f, sumd = 0.0f;;
  int nangle = 0;
  int nm, mn;
  float va[3], vb[3], temp, c1, c2=0.0f, vlo, qe = 0.0f, siga = 0.0f, sigd = 0.0f;
  char poly_col_p[40];
  char poly_col_v[40];

  strcpy(poly_col_p,drvui->poly_col[polyno]);
  strcpy(poly_col_v,poly_col_p);
  Transform_VRML_Color(poly_col_v);
  Transform_POV_Color(poly_col_p);

  if (!(no = (int *) zalloc ((unsigned) ((4 * numb_list) * sizeof (int))))) {
    Error_Box("Unable to get no allocation!");
    return;
  }
  if (!(ns = (int *) zalloc ((unsigned) ((4 * numb_list) * sizeof (int))))) {
    Error_Box("Unable to get ns allocation!");
    free (no);
    return;
  }
  radius = drvui->rad_edge * 0.005f *Scale;
  memset(no,0,sizeof(no));
  memset(ns,0,sizeof(ns));

  if (drvui->poly_rad_edge[polyno] != 0.)
    radius = drvui->poly_rad_edge[polyno] * 0.005f *Scale;

  Vol = 0.0f;
  for (i=0; i<numb_list-2; i++) {
      ih[0] = i;
      for (j=i+1; j<numb_list-1; j++) {
          ih[1] = j;
          for (k=j+1; k<numb_list; k++) {
              ih[2] = k;
              for (m=0; m<3; m++) {
                  v1[m] = s_vert[3*vertex_list[j]+m] - s_vert[3*vertex_list[i]+m];
                  v2[m] = s_vert[3*vertex_list[k]+m] - s_vert[3*vertex_list[i]+m];
                  v4[m] = s_vert[3*vertex_list[i]+m] - cx[m];
              }
              vcross(v1,v2,v3);
              Area = 0.5f * vlength(v3);
              Z0 = 0.5f * vdot(v3,v4)/Area;
              if (fabs(Z0) < 0.0001) continue;     /* 3 points are in a line */
              for (l=0; l<numb_list; l++) {
                  if (l == i || l == j || l == k) continue;
                  for (m=0; m<3; m++) {
                      v0[m] = s_vert[3*vertex_list[i]+m] - s_vert[3*vertex_list[l]+m];
                  }
                  Z = 0.5f * vdot(v0, v3);
                  if (Z * Z0 < 0.0f) continue;      /* not all points on same side */
                  for (m=0; m<3; m++) {
                      v3[m] *= 0.5f/Area;           /* directions cosines of plane normal */
                  }
              }
              Vol += Area * (float)(fabs(Z0)/3.0);
              for (l=0; l<2; l++) {
                  nm = ih[l];
                  for (m=l+1; m<3; m++) {
                      mn = ih[m];
                      for (n=0; n<3; n++) {
                          va[n] = s_vert[3*vertex_list[nm]+n] - cx[n];
                          vb[n] = s_vert[3*vertex_list[mn]+n] - cx[n];
                      }
                      temp = vdot(va, vb)/(vlength(va) * vlength(vb));
                      temp = min(temp, 1.0f);
                      temp = max(-1.0f,temp);
                      temp = (float)(atan2(sqrt(1.0f - temp*temp),temp) * 180.0/PI);
                      sumth += temp;
                      sumth2 += temp*temp;
                      nangle++;
                  }
              }
          }
      }
  }
  for (i=0; i<numb_list; i++) {
      for (n=0; n<3; n++) {
          va[n] = s_vert[3*vertex_list[i]+n] - cx[n];
      }
      temp = vlength(va);
      sumd2 += temp * temp;
      sumd += temp;

  }
  if (numb_list == 4) {
      c1 = 9.0f * (float)(sqrt(3.0)/8.0);
      c2 = 109.045f;
  } else if (numb_list == 6) {
      c1 = 0.75f;
      c2 = 90.0f;
  } else {
      c1 = 0.0f;
  }
  if (c1) {
      vlo = (float)(exp(2.0 * log(c1 * Vol)/3.0));
      qe = sumd2 / ((float)numb_list * vlo);
      nangle = (nangle + 1)/2;
      sumth2 *= 0.5f;
      sumth *= 0.5f;
      siga = (sumth2 - 2.0f * c2 * sumth + nangle * c2 * c2)/(float)(nangle - 1);
      sigd = (sumd2 - sumd * sumd / (float)numb_list)/(float)(nangle - 1);
  }
  for (i = 0; i < numb_list - 2; ++i) {
    no[0] = vertex_list[i];             /* first point */
    if (DepthCue != 0.) {
/*calculate z coordinate after rotation, and scale edge width accordingly*/
      radius = 0.0f;
      for (l = 0; l <= 2; l++){
        radius += (float) G_Rot[l][2] * s_vert[3 * no[0] + l];
      }
      radius = drvui->rad_edge * 0.005f * Scale + DepthCue * radius;
    }
    for (m = 0; m <= 2; ++m)
      p[m] = s_vert[3 * no[0] + m];
    for (j = i + 1; j < numb_list - 1; ++j) {
      no[1] = vertex_list[j];             /* second point */
      for (k = 0; k <= 2; ++k)
        v1[k] = s_vert[3 * no[1] + k] - p[k];
      for (k = j + 1; k < numb_list; ++k) {
        if (numb_list == 4) {             /* tetrahedron */
          poly_list[draw_list++] = 3;         /* number of vertices for triangle */
          poly_list[draw_list++] = vertex_list[i] + 1;
          poly_list[draw_list++] = vertex_list[j] + 1;
          poly_list[draw_list++] = vertex_list[k] + 1;
          if (draw_list > 4 * NvertM) {
            poly_list = (int *)zalloc ((unsigned) (4 * (nvert+2) * sizeof (int)));
            if (!warn) {
              fprintf (drvui->flout,"Overrun of poly_list.\n");
              fprintf (drvui->flout, "Overrun of poly_list.\n");
              Error_Box("Overrun of poly_list. Please send 'str' file\n"
                   "to Larry.Finger@@lwfinger.net.");
              warn = 1;
            }
          }

          imk = vertex_list[i];
          jmk = vertex_list[j];
          kmk = vertex_list[k];
          for (m=0; m<3; m++) {
              v1[m] = s_vert[3*imk+m] - s_vert[3*jmk+m];
              v2[m] = s_vert[3*imk+m] - s_vert[3*kmk+m];
              v3[m] = s_vert[3*kmk+m] - s_vert[3*jmk+m];
          }
          d1 = vlength(v1);
          d2 = vlength(v2);
          d3 = vlength(v3);
          if ((d1 > 0.01) && (d2 > 0.01) && (d3 > 0.01)) {
            fprintf (drvui->fpoutp,
                 " triangle{< %8.5f, %8.5f, %8.5f>,\n",
                 s_vert[3 * imk], s_vert[3 * imk + 1],
                 s_vert[3 * imk + 2]);
            fprintf (drvui->fpoutp,
                 "<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>  texture{pigment{color %s  }}\n",
                 s_vert[3 * jmk], s_vert[3 * jmk + 1],
                 s_vert[3 * jmk + 2], s_vert[3 * kmk],
                 s_vert[3 * kmk + 1], s_vert[3 * kmk + 2],
                 poly_col_p);
            fprintf(drvui->fpoutp," }\n");
            glPushMatrix ();
            glLoadName (i0*100000+i0);
            glPushName (i1);
            (void)sscanf (poly_col_v, "%f %f %f", &glr, &glg, &glb);
            glColor3f (glr, glg, glb);
            glDisable(GL_LIGHTING);
            glBegin (GL_TRIANGLES);
            glVertex3f (s_vert[3 * imk], s_vert[3 * imk + 1], s_vert[3 * imk + 2]);
            glVertex3f (s_vert[3 * jmk], s_vert[3 * jmk + 1], s_vert[3 * jmk + 2]);
            glVertex3f (s_vert[3 * kmk], s_vert[3 * kmk + 1], s_vert[3 * kmk + 2]);
            glEnd ();
            glEnable(GL_LIGHTING);
            glPopName();
            glPopMatrix ();
            if (edges) {
              float df[3], at[3];
              int m;
              char color[128];
              if (drvui->poly_rad_edge[polyno] > 0.005f)
                strncpy (edgecolor, drvui->poly_col_edge[polyno],39);
              else
                strncpy (edgecolor, drvui->col_edge,39);
              trim_string(edgecolor,40);
              if (!strlen(edgecolor)) strcpy(edgecolor,"White");
              strcpy(color,edgecolor);
              Transform_VRML_Color(color);
              Transform_POV_Color(edgecolor);
              fprintf (drvui->fpoutp,
                   " cylinder { <%f,%f,%f>,<%f,%f,%f>,%f\n",
                   s_vert[3 * imk], s_vert[3 * imk + 1],
                   s_vert[3 * imk + 2], s_vert[3 * jmk],
                   s_vert[3 * jmk + 1], s_vert[3 * jmk + 2], radius);
              fprintf (drvui->fpoutp,
                   "  texture{pigment{color %s  }}\n", edgecolor);
              fprintf(drvui->fpoutp," }\n");
              for (m=0; m<3; m++) {
                at[m] = s_vert[3 * imk + m];
                df[m] = s_vert[3 * jmk + m] - s_vert[3 * imk + m];
              }
              push_cylinder(df, at, radius, color); 
              fprintf (drvui->fpoutp,
                   " cylinder { <%f,%f,%f>,<%f,%f,%f>,%f\n",
                   s_vert[3 * jmk], s_vert[3 * jmk + 1],
                   s_vert[3 * jmk + 2], s_vert[3 * kmk],
                   s_vert[3 * kmk + 1], s_vert[3 * kmk + 2], radius);
              fprintf (drvui->fpoutp,
                   "  texture{pigment{color %s  }}\n", edgecolor);
              fprintf(drvui->fpoutp," }\n");
              for (m=0; m<3; m++) {
                at[m] = s_vert[3 * jmk + m];
                df[m] = s_vert[3 * kmk + m] - s_vert[3 * jmk + m];
              }
              push_cylinder(df, at, radius, color); 
              fprintf (drvui->fpoutp,
                   " cylinder { <%f,%f,%f>,<%f,%f,%f>,%f \n",
                   s_vert[3 * kmk], s_vert[3 * kmk + 1],
                   s_vert[3 * kmk + 2], s_vert[3 * imk],
                   s_vert[3 * imk + 1], s_vert[3 * imk + 2], radius);
              fprintf (drvui->fpoutp,
                   "  texture{pigment{color %s  }}\n", edgecolor);
              fprintf(drvui->fpoutp," }\n");
              for (m=0; m<3; m++) {
                at[m] = s_vert[3 * kmk + m];
                df[m] = s_vert[3 * imk + m] - s_vert[3 * kmk + m];
              }
              push_cylinder(df, at, radius, color); 
            }
          }
        } else {
          no[2] = vertex_list[k];         /* third point */
          jplane = 3;
          for (l = 0; l <= 2; ++l)
            v2[l] = s_vert[3 * no[2] + l] - p[l];    /* get v2 */
          sign_z = 0.0f;
          ind = 0;
          v3[0] = v1[1] * v2[2] - v1[2] * v2[1];
          v3[1] = v1[2] * v2[0] - v1[0] * v2[2];
          v3[2] = v1[0] * v2[1] - v1[1] * v2[0];
          for (l = 0; l < numb_list; ++l) {
            vol = 0.0f;
            for (m = 0; m <= 2; ++m)
              vol = vol + (s_vert[3 * vertex_list[l] + m] - p[m]) * v3[m];
            if ((float) fabs (vol) < drvui->polylimit) {
              if (l > k) {
                no[jplane++] = vertex_list[l];     /* add to plane list */
              } else {
                if ((l != i) && (l != j) && (l != k)) {
                  ind = 1;             /* we already have this plane */
                  l = numb_list;         /* skip rest of loop */
                }
              }
            } else {
/* point is not in plane - check if first out-of-plane point found */
              if (sign_z == 0.0) {
                sign_z = vol;             /* yes - set sign */
              } else {
/* if here, not first out of plane - check that vol and sign_z have same sign. 
   If not, set ind=1 to indicate not a plane */
                if (vol * sign_z < 0.0) {
                  ind = 1;
                  l = numb_list;         /* skip rest of inner loop */
                }
              }
            }
          }                	 /* end of loop on l */
          if (ind == 0) {             /* 0 means a surface plane */
            poly_list[draw_list++] = jplane;     /* number of vertices in plane */
            for (l = 0; l < jplane; ++l)
              poly_list[draw_list++] = no[l] + 1;
            fprintf (drvui->fpoutp, "/* %d-sided polygon */\n", jplane);
/* Make vertices consecutive - find center of polygon */
            xm = ym = zm = 0.0f;
            for (ii = 0; ii < jplane; ii++) {
              xm = xm + s_vert[3 * no[ii]];
              ym = ym + s_vert[3 * no[ii] + 1];
              zm = zm + s_vert[3 * no[ii] + 2];
            }
            xm = xm / (float) jplane;
            ym = ym / (float) jplane;
            zm = zm / (float) jplane;

/* sort vertices in consecutive order based on minimal angles between vectors */

            for (ii = 0; ii < jplane; ii++)
              ns[ii] = 0;
            ns[0] = no[0];
            for (ii = 0; ii < jplane - 1; ii++) {
              phi0 = 1000.0f;
              vxs = s_vert[3 * ns[ii]] - xm;
              vys = s_vert[3 * ns[ii] + 1] - ym;
              vzs = s_vert[3 * ns[ii] + 2] - zm;
              bms = (float) sqrt (vxs * vxs + vys * vys + vzs * vzs);
              for (jj = 0; jj < jplane; jj++) {
                for (ll = 0; ll < ii; ll++) {
                  if (ns[ll] == no[jj])
                    jj++;
                  if (jj >= jplane)
                    break;
                }
                if (jj < jplane) {
                  if (ns[ii] != no[jj]) {
                    vx = s_vert[3 * no[jj]] - xm;
                    vy = s_vert[3 * no[jj] + 1] - ym;
                    vz = s_vert[3 * no[jj] + 2] - zm;
                    bm = (float) sqrt (vx * vx + vy * vy + vz * vz);
                    cosphi = (vxs * vx + vys * vy + vzs * vz) / (bms * bm);
                    if (cosphi < -1.) cosphi = -1.;
                    if (cosphi > 1.) cosphi = 1.;
                    phi = acos (cosphi) * RAD;
                    if (phi < phi0) {
                      phi0 = (float) phi;
                      ns[ii + 1] = no[jj];
                    }
                  }
                }
              }
            }
            if (numb_list != 4) {
              draw_list = draw_list - jplane;
              for (l = 0; l < jplane; ++l)
                poly_list[draw_list++] = ns[l] + 1;
            }

            glPushMatrix ();
            glLoadName (i0*100000+i0);
            glPushName (i1);
            (void)sscanf (poly_col_v, "%f %f %f", &glr, &glg, &glb);
            glDisable (GL_LIGHTING);
            glColor3f (glr, glg, glb);
            glBegin (GL_TRIANGLES);

            for (l = 0; l < jplane - 2; ++l) {
              d1 = (float) fabs (s_vert[3 * ns[0]] - s_vert[3 * ns[l + 1]]) +
                   (float) fabs (s_vert[3 * ns[0] + 1] -  s_vert[3 * ns[l + 1] + 1]) +
                   (float) fabs (s_vert[3 * ns[0] + 2] -  s_vert[3 * ns[l + 1] + 2]);
              d2 = (float) fabs (s_vert[3 * ns[0]] - s_vert[3 * ns[l + 2]]) +
                   (float) fabs (s_vert[3 * ns[0] + 1] -  s_vert[3 * ns[l + 2] + 1]) +
                   (float) fabs (s_vert[3 * ns[0] + 2] -  s_vert[3 * ns[l + 2] + 2]);
              d3 = (float) fabs (s_vert[3 * ns[l + 2]] - s_vert[3 * ns[l + 1]]) +
                   (float) fabs (s_vert[3 * ns[l + 2] + 1] -  s_vert[3 * ns[l + 1] + 1]) +
                   (float) fabs (s_vert[3 * ns[l + 2] + 2] -  s_vert[3 * ns[l + 1] + 2]);
              if ((d1 > 0.01) && (d2 > 0.01) && (d3 > 0.01)) {
                fprintf (drvui->fpoutp, " triangle {< %8.5f, %8.5f, %8.5f>,\n",s_vert[3 * ns[0]],
                      s_vert[3 * ns[0] + 1], s_vert[3 * ns[0] + 2]);
                fprintf (drvui->fpoutp, "<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
                      s_vert[3 * ns[l + 1]],s_vert[3 * ns[l + 1] + 1],
                      s_vert[3 * ns[l + 1] + 2],s_vert[3 * ns[l + 2]],
                      s_vert[3 * ns[l + 2] + 1],s_vert[3 * ns[l + 2] + 2]);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n",poly_col_p);
                fprintf(drvui->fpoutp," }\n");

/* 
      need to add something like
      glNormal3f(xm-xc,ym-yc,zm-zc ); 
      for lighting -  xc,yc,zc being the center of the polyhedron
      (which is not normally available to build_poly_list)
*/
                glVertex3f (s_vert[3 * ns[0]],s_vert[3 * ns[0] + 1], s_vert[3 * ns[0] + 2]);
                glVertex3f (s_vert[3 * ns[l + 1]],s_vert[3 * ns[l + 1] + 1],s_vert[3 * ns[l + 1] + 2]);
                glVertex3f (s_vert[3 * ns[l + 2]],s_vert[3 * ns[l + 2] + 1],s_vert[3 * ns[l + 2] + 2]);
              }
            }
            glEnd ();
            glPopName();
            glEnable (GL_LIGHTING);
            glPopMatrix ();
            if (edges) {
              float at[3], df[3];
              int m;
              char color[128];
              if (drvui->poly_rad_edge[polyno] >= 0.001f)
                strncpy (edgecolor, drvui->poly_col_edge[polyno],39);
              else
                strncpy (edgecolor, drvui->col_edge,39);
              trim_string(edgecolor,40);
              if (!strlen(edgecolor)) strcpy(edgecolor,"White");
              strcpy(color, edgecolor);
              Transform_VRML_Color(color);
              Transform_POV_Color(edgecolor);
              for (l = 0; l < jplane - 1; ++l) {
                d1 = (float) fabs (s_vert[3 * ns[l]] - s_vert[3 * ns[l + 1]]) +
                     (float) fabs (s_vert[3 * ns[l] + 1] - s_vert[3 * ns[l + 1] + 1]) +
                     (float) fabs (s_vert[3 * ns[l] + 2] - s_vert[3 * ns[l + 1] + 2]);

                if (d1 > 0.01) {
                  fprintf (drvui->fpoutp," cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                     s_vert[3 * ns[l]],s_vert[3 * ns[l] + 1],s_vert[3 * ns[l] + 2],
                     s_vert[3 * ns[l + 1]],s_vert[3 * ns[l + 1] + 1],s_vert[3 * ns[l + 1] + 2], radius);
                  fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", edgecolor);
                  fprintf(drvui->fpoutp," }\n");
                  for (m=0; m<3; m++) {
                    at[m] = s_vert[3 * ns[l] + m];
                    df[m] = s_vert[3 * ns[l + 1] + m] - at[m];
                  }
                  push_cylinder(df, at, radius, color);
                }
              }
              d1 = (float) fabs (s_vert[3 * ns[0]] - s_vert[3 * ns[jplane - 1]]) +
                   (float) fabs (s_vert[3 * ns[0] + 1] -  s_vert[3 * ns[jplane - 1] + 1]) +
                   (float) fabs (s_vert[3 * ns[0] + 2] -  s_vert[3 * ns[jplane - 1] + 2]);
              if (d1 > 0.01) {
                fprintf (drvui->fpoutp, " cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                   s_vert[3 * ns[0]],s_vert[3 * ns[0] + 1],s_vert[3 * ns[0] + 2],
                   s_vert[3 * ns[jplane - 1]],s_vert[3 * ns[jplane - 1] + 1],
                   s_vert[3 * ns[jplane - 1] + 2], radius);
                fprintf (drvui->fpoutp, "  texture{pigment{color %s  }}\n", edgecolor);
                fprintf(drvui->fpoutp," }\n");
                for (m=0; m<3; m++) {
                  at[m] = s_vert[3 * ns[0] + m];
                  df[m] = s_vert[3 * ns[jplane - 1] + m] - at[m];
                }
                push_cylinder(df, at, radius, color);
              }
            }
          }
        }                     /* end of numb_list != 4 */
      }                    	 /* end of for (k.... */
    }
  }
  if (c1) {
      fprintf(drvui->flout," %7.3f %8.5f %8.5f %8.3f\n",Vol, sigd, qe, siga);
  } else {
      fprintf(drvui->flout," %7.3f %8.5f\n",Vol, sigd);
  }
  free (no);
  free (ns);
  return;
}                    	 /* end of build_poly_list */

/* ************************************************************** */
/* ************************************************************** */

void draw_cell (int docell)

/* procedure to add unit-cell outline to geometry file */
/* docell - non-zero if unit-cell to be drawn */
{
    int i,j,k;
    float vert[3];
    float radius;
//  static char abc[] = "cbao";
    float df[3];
    float at[3];
    char col_cell_v[40];
    char col_cell_p[40];
    int nocell;

    radius = rad_cell * 0.01f *Scale;
    if (!drvui->no_subsys)
        drvui->no_subsys = 1;

    for (nocell=0; nocell<drvui->no_subsys; nocell++) {
        nvert = 0;                     /* initialize vertex list */
        for (i = 0; i <= 2; ++i)
            vert[i] = 0.0f;
        sub_add_vert_nc (vert, nocell);             // place 0,0,0 in list
        vert[2] = 1.0f;
        sub_add_vert_nc (vert, nocell);             // place 0,0,1 in list
        vert[1] = 1.0f;
        sub_add_vert_nc (vert, nocell);             // place 0,1,1 in list
        vert[2] = 0.0f;
        sub_add_vert_nc (vert, nocell);             // place 0,1,0 in list
        vert[1] = 0.0f;
        sub_add_vert_nc (vert, nocell);             // place 0,0,0 in list
        vert[0] = 1.0f;
        sub_add_vert_nc (vert, nocell);             // place 1,0,0 in list
        vert[1] = 1.0f;
        sub_add_vert_nc (vert, nocell);             // place 1,1,0 in list
        vert[2] = 1.0f;
        sub_add_vert_nc (vert, nocell);             // place 1,1,1 in list
        vert[1] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,0,1 in list */
        vert[2] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,0,0 in list */
        vert[1] = 1.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,1,0 in list */
        vert[0] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 0,1,0 in list */
        vert[2] = 1.0f;
        sub_add_vert_nc (vert, nocell);             /* place 0,1,1 in list */
        vert[0] = 1.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,1,1 in list */
        vert[1] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,0,1 in list */
        vert[0] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 0,0,1 in list */

        if (docell) {
            strcpy(col_cell_v,drvui->col_cell);
            Transform_VRML_Color(col_cell_v);
            strcpy(col_cell_p,drvui->col_cell);
            Transform_POV_Color(col_cell_p);
            if (Vrml2) {
                fprintf (drvui->fpoutv, " Shape {\n");
                fprintf (drvui->fpoutv,"  geometry IndexedLineSet{ coord Coordinate{ point[\n");
            } else {
                fprintf (drvui->fpoutv, " Separator{\n");
                fprintf (drvui->fpoutv, "  Material{\n");
                fprintf (drvui->fpoutv, "   diffuseColor %s\n", col_cell_v);
                fprintf (drvui->fpoutv, "  }\n");
                fprintf (drvui->fpoutv, "  Coordinate3{ point[\n");
            }
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[0], s_vert[1],s_vert[2]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[3], s_vert[4],s_vert[5]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[6], s_vert[7],s_vert[8]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[9], s_vert[10],s_vert[11]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[15], s_vert[16],s_vert[17]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[18], s_vert[19],s_vert[20]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[21], s_vert[22],s_vert[23]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f]\n  }\n", s_vert[24],s_vert[25], s_vert[26]);
            if (Vrml2) {
                fprintf (drvui->fpoutv, "  coordIndex[");
            } else {
                fprintf (drvui->fpoutv, "  IndexedLineSet{coordIndex[");
            }
            fprintf (drvui->fpoutv, "0,1,2,3,0,4,5,6,7,1,2,6,5,3,0,4,7,-1]\n");
            if (Vrml2) {
                fprintf (drvui->fpoutv, "  color Color { color [%s]}\n  colorIndex[0]\n"
                   "  colorPerVertex FALSE\n",col_cell_v);
            }
            fprintf (drvui->fpoutv, " }\n");
            fprintf (drvui->fpoutv, "}\n");

            for (i=0;i<12;i++) {                 // generate unit-cell edges for openGL
                j = i;
                if (i == 9) j = 10;
                if (i == 10) j = 12;
                k = j+1;
                if (i == 11) {
                    j = 8;
                    k = 1;
                }
                at[0] = s_vert[3*j];              // beginning vertex
                at[1] = s_vert[3*j+1];
                at[2] = s_vert[3*j+2];
                df[0] = s_vert[3*k] - at[0];      // difference from start to end
                df[1] = s_vert[3*k+1] - at[1];
                df[2] = s_vert[3*k+2] - at[2];
                push_cylinder(df,at, radius, col_cell_v);  // draw the cylinder
            }
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",
                    s_vert[0], s_vert[1], s_vert[2], s_vert[3], s_vert[4],
                    s_vert[5], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[3], s_vert[4], s_vert[5], s_vert[6], s_vert[7],
                    s_vert[8], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[6], s_vert[7], s_vert[8], s_vert[9], s_vert[10],
                    s_vert[11], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>, %f\n",
                    s_vert[9], s_vert[10], s_vert[11], s_vert[12], s_vert[13],
                    s_vert[14], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[12], s_vert[13], s_vert[14], s_vert[15], s_vert[16],
                    s_vert[17], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[15], s_vert[16], s_vert[17], s_vert[18], s_vert[19],
                    s_vert[20], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[18], s_vert[19], s_vert[20], s_vert[21], s_vert[22],
                    s_vert[23], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[21], s_vert[22], s_vert[23], s_vert[24], s_vert[25],
                    s_vert[26], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[24], s_vert[25], s_vert[26], s_vert[27], s_vert[28],
                    s_vert[29], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[30], s_vert[31], s_vert[32], s_vert[33], s_vert[34],
                    s_vert[35], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[36], s_vert[37], s_vert[38], s_vert[39], s_vert[40],
                    s_vert[41], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[24], s_vert[25], s_vert[26], s_vert[3], s_vert[4],
                    s_vert[5], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
        }     /* if (docell) */
    }         /* for(nocell  */

    if (Display_axes) {

        nvert = 0;                     /* initialize vertex list */
        generate_triple();
        if (radius == 0.) radius = 0.0002f *Scale;
        strcpy(col_cell_v,drvui->col_cell);
        Transform_VRML_Color(col_cell_v);
        strcpy(col_cell_p,drvui->col_cell);
        Transform_POV_Color(col_cell_p);

/* draw vector triple in POV */
        fprintf (drvui->fpoutp, "/* drawing vector triple */");
        fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",s_vert[0], s_vert[1], s_vert[2], 
              s_vert[3], s_vert[4],s_vert[5], radius);
        fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
        fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
        fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",s_vert[0], s_vert[1], s_vert[2],
              s_vert[6], s_vert[7],s_vert[8], radius);
        fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
        fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
        fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",s_vert[0], s_vert[1], s_vert[2],
              s_vert[9], s_vert[10],s_vert[11], radius);
        fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
        fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
/* draw vector triple in VRML */
        if (Vrml2) {
            fprintf (drvui->fpoutv, " Shape{\n");
            fprintf (drvui->fpoutv, "  geometry IndexedLineSet { coord Coordinate{ point [\n");
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[0], s_vert[1],s_vert[2]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[3], s_vert[4],s_vert[5]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[6], s_vert[7],s_vert[8]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f]\n  }\n", s_vert[9], s_vert[10], s_vert[11]);
            fprintf (drvui->fpoutv, "  coordIndex [");
            fprintf (drvui->fpoutv, "0,3,0,2,0,1,-1]\n");
            fprintf (drvui->fpoutv, "   color Color { color [%s]}\n   colorIndex[0]\n"
                    "   colorPerVertex FALSE\n",col_cell_v);
            fprintf (drvui->fpoutv, "  }\n");
            fprintf (drvui->fpoutv, " }\n");
        } else {
            fprintf (drvui->fpoutv, " Separator{\n");
            fprintf (drvui->fpoutv, "  Material{\n");
            fprintf (drvui->fpoutv, "   diffuseColor %s\n", col_cell_v);
            fprintf (drvui->fpoutv, "  }\n");
            fprintf (drvui->fpoutv, "  Coordinate3{ point[\n");
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[0], s_vert[1],s_vert[2]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[3], s_vert[4],s_vert[5]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[6], s_vert[7],s_vert[8]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f]\n  }\n", s_vert[9], s_vert[10], s_vert[11]);
            fprintf (drvui->fpoutv, "  IndexedLineSet{coordIndex[");
            fprintf (drvui->fpoutv, "0,3,0,2,0,1,-1]\n");
            fprintf (drvui->fpoutv, "  }\n");
            fprintf (drvui->fpoutv, " }\n");
        }
    }

    nvert = 0;                     /* empty vertex list */
}                    	         /* end of draw_cell */

/* ************************************************************** */
/* ************************************************************** */

void draw_GL_triple(void)
{
// draw the lines for the vector triple in openGL
    int i;
    float at[3], df[3];
    float radius = rad_cell * 0.01f *Scale;
    char col_cell_v[40];

    strcpy(col_cell_v,drvui->col_cell);
    Transform_VRML_Color(col_cell_v);

    if (radius ==0.) radius = 0.0002f *Scale;
    
    for (i=0; i<3; i++) {
        at[i] = s_vert[i];
        df[i] = s_vert[3 + i] - at[i];
    }
    push_cylinder(df, at, radius, col_cell_v);
    for (i=0; i<3; i++) {
        at[i] = s_vert[i];
        df[i] = s_vert[6 + i] - at[i];
    }
    push_cylinder(df, at, radius, col_cell_v);
    for (i=0; i<3; i++) {
        at[i] = s_vert[i];
        df[i] = s_vert[9 + i] - at[i];
    }
    push_cylinder(df, at, radius, col_cell_v);
}

/* ************************************************************** */
/* ************************************************************** */

void generate_spheres (void)

/* routine to generate spherical objects and add then to edit list */
{
  float *radii;                     /* dynamic array for sphere radii */
  int Sphere_Count;                 /* Counter for number output */
  int i, j, k;                      /* loop counters */

  Sphere_Count = 0;
  for (i = 1; i < drvui->nsphere; ++i) {         /* loop through spheres */
      if (drvui->sphere_fn[i] != drvui->frame_no) continue;  // skip if not in this frame
      nvert = 0;                     /* clear the vertex list */
      for (j = 0; j < natom; ++j) {  /* loop through atoms */
          if (drvui->atom_fn[j] != drvui->frame_no) continue;
          if ((drvui->atom_n[j] & 255) == i)
              find_all_in_box (j);
      }
      if (nvert > 0) {                 /* if spheres present */
/* allocate storage for radii of spheres */
          if (!(radii = (float *) zalloc ((unsigned) (nvert * sizeof (float))))) {
              Error_Box("Unable to allocate memory for storage of spheres.");
              return;
          };
          for (k = 0; k < nvert; ++k) {         /* copy radii info to storage */
              radii[k] = drvui->sphere_size[i] * drvui->Sphere_Mult;
              if (drvui->sphere_size[i] <0.)
                radii[k] *= - drvui->vert_occ[k];
          }
          Output_Spheres (radii, i);
          free (radii);
      }
      Sphere_Count += nvert;
  }
  if (Sphere_Count > 0) {
      fprintf (drvui->fcns,"%4d spheres.\n", Sphere_Count);
      fprintf (drvui->flout, "Generated %4d spheres.\n", Sphere_Count);
  }
}                    	 /* end of generate_spheres */

/* ************************************************************** */
/* ************************************************************** */

void generate_triple(void)
{
// generate the vertices for the vector triple
    
    int i, j;
    float vert[3];
   
    nvert = 0;
    for (i = 0; i < 3; i++)
      vert[i] = 0.0f;
    add_vert_nc (vert);                 /* place origin at 0,0,0 in list */
    vert[2] = 1.0f / drvui->lat_con[2];
    add_vert_nc (vert);                 /* place 1 A vector along 001 */
    vert[2] = 0.0f;
    vert[1] = 1.0f / drvui->lat_con[1];
    add_vert_nc (vert);                 /* place 1 A vector along 010 */
    vert[1] = 0.0f;
    vert[0] = 1.0f / drvui->lat_con[0];
    add_vert_nc (vert);                 /* place 1 A vector along 100 */
    vert[2] = 1.4f / drvui->lat_con[2];
    vert[0] = 0.0f;
    add_vert_nc (vert);                 /* place 1.4 A vector along 001 */
    vert[2] = 0.0f;
    vert[1] = 1.4f / drvui->lat_con[1];
    add_vert_nc (vert);                 /* place 1.4 A vector along 010 */
    vert[1] = 0.0f;
    vert[0] = 1.4f / drvui->lat_con[0];
    add_vert_nc (vert);                 /* place 1.4 A vector along 100 */
    for (i=nvert-1; i>=0; --i) {             /* Translate coordinates */
      for (j = 0; j < 3; j++) {
        s_vert[3 * i + j] += offset[j] - s_vert[j];
      }
    }
}

/* ************************************************************** */
/* ************************************************************** */

void push_cylinder (float df[3], float at[3], float radius, char * color)

// routine to push an openGL cylinder that represents a unit-cell edge
{
    float beta, gamma;
    float glr,glg,glb;
    float d; //,radius;
    GLUquadricObj *glu_quadObj;

//    radius = radius * 0.01f *Scale;
    d = (float)sqrt(df[0]*df[0]+df[1]*df[1]+df[2]*df[2]);
    beta = (float) atan2 (df[0], df[1] + 0.0000001f);    /* rotation angle about Z (in radians) */
    gamma = (float) sqrt (df[0] * df[0] + df[1] * df[1]);
    gamma = (float) atan2 (gamma, df[2]);    /* Rotation angle about X */
    glu_quadObj = gluNewQuadric ();
    glPushMatrix ();
    (void)sscanf (color, "%f %f %f", &glr, &glg, &glb);
    glColor3f (glr, glg, glb);
    glTranslatef (at[0], at[1], at[2]);
    glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.0f);
    glRotatef (-gamma * (float) RAD, 1.0f, 0.0f, 0.0f);
    gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
    gluCylinder (glu_quadObj,radius,radius, d, 10, 1);
    gluDeleteQuadric(glu_quadObj);
    glPopMatrix ();
}

/* ************************************************************** */
/* ************************************************************** */

void generate_lsq_planes (void)

// routine to calculate and display a least squares plane through a set of atoms
{
  int i, j, jj, k, l, m;
  float d;
  float glr,glg,glb;
  char thecolor[40];
  float c[3],s[6],vects[3][3],vals[3];
  double vx[50][15],vy[50][15],vz[50][15];
  double vsx[50][15],vsy[50][15],vsz[50][15];
  double t0[3],t1[3],t2[3],t3[3],p0[3],p1[3],p2[3],p3[3];
  float dummy;
  int nsets=0;
  double sx,sy,sz;
  int nx,add,nn[50];
  int numplanes=0;
  double pn0[MAX_PLANE],pn1[MAX_PLANE],pn2[MAX_PLANE];
  double v1n,v2n,dot,angval;

  if (drvui->nbplane==1) return;

  for (m=1;m<drvui->nbplane;m++) {

    nsets = 0;

// first pass through atom list to find maximum number of planes to expect
// (numbers may vary for individual atoms due to view box clipping)

    for (i = 0; i <= natom; ++i) { 

      if (drvui->atom_fn[i] != drvui->frame_no) continue;

      add=0;
      for (j=0;j<drvui->nbatoms[m];j++){ // only if part of l.s. plane
        if (  (check_atom_name(drvui->atom_l[i],drvui->bplane_t[m][j])==1) 
           && ( drvui->sv_atom_n[i]== drvui->bplane_n[m][j] || drvui->bplane_n[m][j]<0) ) 
          add=1;
      }
      if (add==0) continue;
    
      nvert=0;
      find_all_in_box (i); // find all copies of this atom in the display box

      if (nvert ==0) {
        Error_Box("At least one atom of the l.s. plane is not in the current display box");
        return;
      }
      if (nvert>nsets) nsets=nvert;
    }

    if (nsets >50) {
        Error_Box("Too many copies (max. 50) of a best fitting plane.");
        return;
      }

// second pass through master atom list:
// assemble atom lists for each copy of the current plane in the view box

    nx = 0;

    for (i = 0; i <= natom; ++i) {

      if (drvui->atom_fn[i] != drvui->frame_no) continue;

      add=0;
      for (j=0;j<drvui->nbatoms[m];j++){ // only if part of l.s. plane
        if (  (check_atom_name(drvui->atom_l[i],drvui->bplane_t[m][j])==1) 
           && ( drvui->sv_atom_n[i]== drvui->bplane_n[m][j] || drvui->bplane_n[m][j]<0) ) 
          add=1;
      }
      if (add==0) continue;
    
      nvert=0;
      find_all_in_box (i); // find all copies of this atom in the display box

      if (nx==0) { // when lists are empty, add one copy to each
        for (jj=0;jj<nsets;jj++){
          vx[jj][0]=o_vert[3*jj]-origin[0];	
          vy[jj][0]=o_vert[3*jj+1]-origin[1];	
          vz[jj][0]=o_vert[3*jj+2]-origin[2];	
          vsx[jj][0]=s_vert[3*jj];	
          vsy[jj][0]=s_vert[3*jj+1];	
          vsz[jj][0]=s_vert[3*jj+2];
          nn[jj]=1;	
        }
        nx=1;
      } else {
        for (jj=0;jj<nvert;jj++){ // find a list that already contains
          add=0;			// atoms within bonding distance
          for (k=0;k<nsets;k++) {
            for (l=0;l<nx;l++) {
              d=(float)((s_vert[3*jj]-vsx[k][l])*(s_vert[3*jj]-vsx[k][l])		
                 +(s_vert[3*jj+1]-vsy[k][l])*(s_vert[3*jj+1]-vsy[k][l])
                 +(s_vert[3*jj+2]-vsz[k][l])*(s_vert[3*jj+2]-vsz[k][l]));
              if (d >0. && d < 1.6*1.6) {
                vx[k][nx]=o_vert[3*jj]-origin[0];	
                vy[k][nx]=o_vert[3*jj+1]-origin[1];	
                vz[k][nx]=o_vert[3*jj+2]-origin[2];	
                vsx[k][nx]=s_vert[3*jj];	
                vsy[k][nx]=s_vert[3*jj+1];	
                vsz[k][nx]=s_vert[3*jj+2];
                nn[k]++;	
                add=1;
                break;
              }
            }
            if (add==1) continue;
          }
        }
        nx++;
        if (nx >14) {
          Error_Box("Cannot handle more than 15 atoms in a best fitting plane.");
          return;
        }
      }
    } // loop over all atoms i in master list

// calculate and render the best fitting plane for each set of atoms

    for (i=0;i<nsets;i++) { 
      if (nn[i]<drvui->nbatoms[m]) continue; // drop incomplete (clipped) planes

      c[0]=c[1]=c[2]=0.;
      sx=sy=sz=0.;
      s[0]=s[1]=s[2]=s[3]=s[4]=s[5]=0.;
      for (k=0;k<nx;k++) {
        c[0] += (float)vsx[i][k];
        c[1] += (float)vsy[i][k];
        c[2] += (float)vsz[i][k];
        sx+=vx[i][k];
        sy+=vy[i][k];
        sz+=vz[i][k];
      }
      c[0] /= nx;
      c[1] /= nx;
      c[2] /= nx;
      sx /= nx;
      sy /= nx;
      sz /= nx;
      for (k=0;k<nx;k++) {
        s[0] += (float)((sx-vx[i][k])*(sx-vx[i][k]));
        s[1] += (float)((sy-vy[i][k])*(sy-vy[i][k]));
        s[2] += (float)((sz-vz[i][k])*(sz-vz[i][k]));
        s[3] += (float)((sx-vx[i][k])*(sy-vy[i][k]));
        s[4] += (float)((sx-vx[i][k])*(sz-vz[i][k]));
        s[5] += (float)((sy-vy[i][k])*(sz-vz[i][k]));
      }

      eigen (&dummy,s,vals,vects);

      d = (float)sqrt( vects[0][0]*vects[0][0]+vects[0][1]*vects[0][1]+vects[0][2]*vects[0][2]);
      vects[0][0] /= d;
      vects[0][1] /= d;
      vects[0][2] /= d;
      d = (float)sqrt( vects[1][0]*vects[1][0]+vects[1][1]*vects[1][1]+vects[1][2]*vects[1][2]);
      vects[1][0] /= d;
      vects[1][1] /= d;
      vects[1][2] /= d;
      d = (float)sqrt( vects[2][0]*vects[2][0]+vects[2][1]*vects[2][1]+vects[2][2]*vects[2][2]);
      vects[2][0] /= d;
      vects[2][1] /= d;
      vects[2][2] /= d;

      fprintf(drvui->flout,"least squares plane %2d(%2d): center at    %f %f %f\n",m,i+1,c[0],c[1],c[2]);
      fprintf(drvui->flout,"                            eigenvalues  %f %f %f\n",vals[0],vals[1],vals[2]);
      fprintf(drvui->flout,"                            plane normal %f %f %f\n",vects[0][0],vects[0][1],vects[0][2]);
      fprintf(drvui->flout,"                            1st vector   %f %f %f\n",vects[1][0],vects[1][1],vects[1][2]);
      fprintf(drvui->flout,"                            2nd vector   %f %f %f\n",vects[2][0],vects[2][1],vects[2][2]);

      // apply scaling factors to generate initial cornerpoints of polygon 
      t0[0]=c[0]+drvui->bplane_d1[m]*vects[2][0];
      t0[1]=c[1]+drvui->bplane_d1[m]*vects[2][1];
      t0[2]=c[2]+drvui->bplane_d1[m]*vects[2][2];
      t1[0]=c[0]+drvui->bplane_d2[m]*vects[1][0];
      t1[1]=c[1]+drvui->bplane_d2[m]*vects[1][1];
      t1[2]=c[2]+drvui->bplane_d2[m]*vects[1][2];
      t2[0]=c[0]-drvui->bplane_d1[m]*vects[2][0];
      t2[1]=c[1]-drvui->bplane_d1[m]*vects[2][1];
      t2[2]=c[2]-drvui->bplane_d1[m]*vects[2][2];
      t3[0]=c[0]-drvui->bplane_d2[m]*vects[1][0];
      t3[1]=c[1]-drvui->bplane_d2[m]*vects[1][1];
      t3[2]=c[2]-drvui->bplane_d2[m]*vects[1][2];
      // treat these as midpoints on the edges of the final rectangle
      // to make width and height factors behave as expected 
      for (jj=0;jj<3;jj++) {
	p0[jj]=c[jj]+(t0[jj]-t1[jj])/2.;
	p1[jj]=c[jj]+(t1[jj]-t2[jj])/2.;
	p2[jj]=c[jj]+(t2[jj]-t3[jj])/2.;
	p3[jj]=c[jj]+(t3[jj]-t0[jj])/2.;
      }

      strncpy(thecolor,drvui->bplane_col[m],39);
      Transform_VRML_Color(thecolor);
      sscanf(thecolor,"%f %f %f",&glr,&glg,&glb);

      glPushMatrix();
      glDisable(GL_LIGHTING);
      glColor3f (glr, glg, glb);
      glBegin(GL_QUADS);
      glNormal3f(c[0]+vects[0][0],c[1]+vects[0][1],c[2]+vects[0][2]);
      glVertex3f((float)p0[0],(float)p0[1],(float)p0[2]);
      glVertex3f((float)p1[0],(float)p1[1],(float)p1[2]);
      glVertex3f((float)p2[0],(float)p2[1],(float)p2[2]);
      glVertex3f((float)p3[0],(float)p3[1],(float)p3[2]);
      glEnd();
      glEnable(GL_LIGHTING);
      glPopMatrix();

      fprintf(drvui->fpoutp,"/* L.S. Plane */\n");
      fprintf(drvui->fpoutp,"triangle {< %8.5f, %8.5f, %8.5f>,\n",
        c[0],c[1],c[2]);
      fprintf(drvui->fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
	p0[0],p0[1],p0[2],p1[0],p1[1],p1[2]);
      fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n",drvui->bplane_col[m]);
      fprintf(drvui->fpoutp," }\n");
      fprintf(drvui->fpoutp,"triangle {< %8.5f, %8.5f, %8.5f>,\n",
        c[0],c[1],c[2]);
      fprintf(drvui->fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
	p1[0],p1[1],p1[2],p2[0],p2[1],p2[2]);
      fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n",drvui->bplane_col[m]);
      fprintf(drvui->fpoutp," }\n");
      fprintf(drvui->fpoutp,"triangle {< %8.5f, %8.5f, %8.5f>,\n",
        c[0],c[1],c[2]);
      fprintf(drvui->fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
	p2[0],p2[1],p2[2],p3[0],p3[1],p3[2]);
      fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n",drvui->bplane_col[m]);
      fprintf(drvui->fpoutp," }\n");
      fprintf(drvui->fpoutp,"triangle {< %8.5f, %8.5f, %8.5f>,\n",
        c[0],c[1],c[2]);
      fprintf(drvui->fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
 	p3[0],p3[1],p3[2],p0[0],p0[1],p0[2]);
      fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n",drvui->bplane_col[m]);
      fprintf(drvui->fpoutp," }\n");

      fprintf(drvui->fpoutv,"# L.S. Plane\n");
      if (Vrml2) {
        fprintf (drvui->fpoutv, " Shape {");
        fprintf (drvui->fpoutv, "appearance Appearance {\n");
        fprintf (drvui->fpoutv,"   material Material {diffuseColor %s} \n \n",
          thecolor);
      } else {
        fprintf (drvui->fpoutv, " Separator {\n        ");
        fprintf (drvui->fpoutv, "  Material {\n");
        fprintf (drvui->fpoutv, "   diffuseColor %s \n  }\n",thecolor);
      }
      if (Vrml2)
        fprintf (drvui->fpoutv,"}\n geometry IndexedFaceSet { coord Coordinate{ point [\n"); 
      else
        fprintf (drvui->fpoutv, "Coordinate3 { point [\n");
	  
      fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f,\n",
	p0[0],p0[1],p0[2]);
      fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f,\n",
	p1[0],p1[1],p1[2]);
      fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f,\n",
	p2[0],p2[1],p2[2]);
      fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f]\n",
	p3[0],p3[1],p3[2]);
	   
      if (Vrml2) {
        fprintf (drvui->fpoutv, " }\n");
        fprintf (drvui->fpoutv, " coordIndex [0,1,2,3,-1]\n");
        fprintf (drvui->fpoutv, " solid FALSE\n");
        fprintf (drvui->fpoutv, " convex TRUE\n");
        fprintf (drvui->fpoutv, " creaseAngle 1.5708\n");
        fprintf (drvui->fpoutv, "  }\n }\n");
      } else {
        fprintf (drvui->fpoutv, "}\n IndexedFaceSet { coordIndex [0,1,2,3,-1]\n");
        fprintf (drvui->fpoutv,"  }\n }\n");
      }
	if (i==0) {
      pn0[m]=vects[0][0];
      pn1[m]=vects[0][1];
      pn2[m]=vects[0][2];
        }
    numplanes ++;

    } // for all l.s. planes

  } // for all bestplane keywords


  fprintf(drvui->flout,"Angles between planes:\n");
  for (i=1;i<m;i++)  {
    v1n=pn0[i]*pn0[i]+pn1[i]*pn1[i]+pn2[i]*pn2[i];
    v1n=sqrt(v1n);
    for (j=i+1;j<m;j++) {
      v2n=pn0[j]*pn0[j]+pn1[j]*pn1[j]+pn2[j]*pn2[j];
      v2n=sqrt(v2n);
      dot = pn0[i]*pn0[j]+pn1[i]*pn1[j]+pn2[i]*pn2[j];
      angval= RAD * acos(dot /(v1n*v2n));
      fprintf(drvui->flout,"Plane %d - Plane %d: %5.3f\n",i,j,angval);
    }
  }

  fprintf(drvui->flout,"Generated %d best fitting planes.\n",numplanes);


}

